﻿Option Strict On

Imports System.Collections.Generic

Public Class Form1

    Private m_Backbuffer As Image
    Private m_Points As New List(Of Point)

    Private m_Times As New List(Of ULong)
    Private m_Generation As Integer = 0

    Private Const r As Integer = 1

    Private Sub Form1_Load(ByVal sender As Object, _
                           ByVal e As System.EventArgs) Handles Me.Load

        m_Backbuffer = New Bitmap(picOut.Width, picOut.Height)

        Dim Rand As New Random

        Do Until m_Points.Count >= 500
            Dim P = New Point(Rand.Next(80, picOut.Width - 80), _
                              Rand.Next(50, picOut.Height - 80))
            If ((P.X - Width / 2) ^ 2 + (P.Y - Height / 2) ^ 2) < 180 ^ 2 _
                OrElse Rand.NextDouble() < 0.1 Then Call m_Points.Add(P)
        Loop

        Call DrawPoints(m_Points, Color.Orange)
    End Sub

    Private Sub picOut_MouseDown(ByVal sender As System.Object, _
            ByVal e As System.Windows.Forms.MouseEventArgs) _
            Handles picOut.MouseDown

        Call m_Points.Add(e.Location)
        Call DrawPoints(m_Points, Color.Orange, True)
        m_Generation = 0
        lblGeneration.Text = "Eingabe"
    End Sub

    Private Sub picOut_Paint(ByVal sender As System.Object, _
            ByVal e As System.Windows.Forms.PaintEventArgs) Handles picOut.Paint

        Call e.Graphics.DrawImage(m_Backbuffer, New Point(0, 0))
    End Sub

    Private Sub LöschenToolStripMenuItem_Click(ByVal sender As System.Object, _
            ByVal e As System.EventArgs) Handles LöschenToolStripMenuItem.Click

        Call m_Points.Clear()
        Call DrawPoints(m_Points, Color.Orange, True)

        m_Generation = 0
        lblGeneration.Text = "Eingabe"
    End Sub

    Private Sub DrawPoints(ByVal Points As IEnumerable(Of Point), _
                           ByVal Color As Color, _
                           Optional ByVal Clear As Boolean = False)

        Using g = Graphics.FromImage(m_Backbuffer)
            g.SmoothingMode = Drawing2D.SmoothingMode.HighQuality
            If Clear Then g.Clear(Color.White)
            For Each Point In Points
                Using Br = New SolidBrush(Color)
                    Call g.FillEllipse(Br, _
                        New Rectangle(Point.X - r, Point.Y - r, 2 * r, 2 * r))
                End Using
            Next
        End Using
        Call picOut.Invalidate()
    End Sub

    Private Sub BerechnenToolStripMenuItem_Click( _
            ByVal sender As System.Object, _
            ByVal e As System.EventArgs) _
            Handles BerechnenToolStripMenuItem.Click

        Dim Rand As New Random

        If m_Points.Count < 3 Then Exit Sub

        Dim Ball As Circle

        If m_Points.Count > 20 Then

            ' Stimmen initialisieren
            If m_Generation = 0 Then
                Call m_Times.Clear()
                For i = 0 To m_Points.Count - 1
                    Call m_Times.Add(1)
                Next
            End If

            ' Nächste Generation
            m_Generation += 1

            ' Indizes und ausgewählte Punkte speichern
            Dim Indices As New HashSet(Of Integer)
            Dim RandomPoints As New List(Of Point)

            ' Gesamtzahl der Stimmen
            Dim Length = m_Times.Aggregate(Function(s, i) s + i)

            ' 20 zufällige, per Stimmzahl gewichtete Punkte wählen
            Do Until RandomPoints.Count >= 20
                Dim Index = Rand.Next(0, CInt(Length))

                If Not Indices.Contains(Index) Then
                    Dim n = 0UL
                    Dim PointIndex = 0

                    While n <= Index
                        n += m_Times(PointIndex)
                        PointIndex += 1
                    End While

                    Call RandomPoints.Add(m_Points(PointIndex - 1))

                    Call Indices.Add(Index)
                End If
            Loop

            ' Fehler zählen und Stimmen angleichen
            Dim NumMistakes = 0
            Ball = CreateBallNaive(RandomPoints)

            For i = 0 To m_Points.Count - 1
                If Not Ball.Contains(m_Points(i)) Then
                    m_Times(i) *= 2UL
                    NumMistakes += 1
                End If
            Next

            ' Ab hier ist nur noch Ausgabe
            lblMistakes.Text = String.Format("Noch {0} Fehler", NumMistakes)
            lblGeneration.Text = String.Format("{0}. Generation", m_Generation)
            lblRad.Text = String.Format("Radius: {0} ", _
                                        Math.Round(Ball.Radius, 0))

        Else
            Ball = CreateBallNaive(m_Points)
            Call MessageBox.Show("Zu wenig Punkte - Löse direkt", "", _
                                 MessageBoxButtons.OK, _
                                 MessageBoxIcon.Information)
        End If

        Using g = Graphics.FromImage(m_Backbuffer)
            g.SmoothingMode = Drawing2D.SmoothingMode.HighQuality

            Call g.Clear(Color.White)
            For Each Point In m_Points
                Call g.FillEllipse( _
                    If(Ball.Contains(Point), Brushes.Green, Brushes.Red), _
                    New Rectangle(Point.X - r, Point.Y - r, 2 * r, 2 * r))
            Next

            Call Ball.Draw(g)
        End Using

        Call picOut.Invalidate()
    End Sub

    ' Naiver Brute Force-Algorithmus zum Finden des Bounding Balls - O(n^4)
    Private Function CreateBallNaive(ByVal Points As IList(Of Point)) As Circle
        Dim Changed As Boolean
        Dim Best As New Circle(Points(0), Points(1), Points(2))

        ' Alle sinnvollen 3-Punkte-Kombinationen finden, Kreis errechnen und 
        '  schauen, welche die Beste ist
        For i = 0 To Points.Count - 1
            For j = i + 1 To Points.Count - 1
                For k = j + 1 To Points.Count - 1
                    Dim ContainsAll As Boolean = True
                    Dim Current = New Circle(Points(i), Points(j), Points(k))

                    For l = 0 To Points.Count - 1
                        Dim Pnt = Points(l)
                        If (Pnt <> Points(i)) AndAlso _
                        (Pnt <> Points(j)) AndAlso _
                        (Pnt <> Points(k)) AndAlso _
                        (Not Current.Contains(Pnt)) Then
                            ContainsAll = False
                            Exit For
                        End If
                    Next

                    If ContainsAll And _
                            ((Current.Radius < Best.Radius) Or Not Changed) Then
                        Changed = True
                        Best = Current
                    End If
                Next
            Next
        Next
        Return Best
    End Function
End Class
